#pragma once

#include <algorithm>
#include <iostream>
#include <tuple>

#include "common.hh"


namespace blky {

std::vector<uint32_t> PathfindFeatures(
    const std::vector<double>&      probs,
    std::tuple<uint32_t, uint32_t>& block_num,
    uint8_t                         feat_bits,
    uint64_t                        seed) {
  const uint32_t num = std::get<0>(block_num) * std::get<1>(block_num);
  if (num == 0) throw std::runtime_error {"number of blocks is zero"};

  const uint64_t dur = probs.size()/num;
  if (probs.size() == 0) {
    throw std::runtime_error {"size of probability matrix is empty"};
  }
  if (dur*num != probs.size()) {
    throw std::runtime_error {"size of probability matrix is mismatch"};
  }

  blky_pathfinder_t pf = {};
  pf.block_num = num;
  pf.feat_bits = feat_bits;
  pf.seed      = seed;

  blky_pathfinder_init(&pf);

  for (uint64_t t = 0; t < dur; ++t) {
    blky_pathfinder_feed(&pf, probs.data()+t*num);
  }
  assert(pf.step_last);

  uint32_t max_idx  = 0;
  double   max_prob = -1;
  for (uint32_t i = 0; i < num; ++i) {
    if (max_prob < pf.probs[i]) {
      max_prob = pf.probs[i];
      max_idx  = i;
    }
  }

  std::vector<uint32_t> ret;
  blky_pathfinder_step_t* step = pf.step_last;
  uint32_t idx = max_idx;
  for (;;) {
    ret.push_back(idx);
    if (!step) break;
    idx  = step->indices[idx];
    step = step->prev;
  }
  std::reverse(ret.begin(), ret.end());

  blky_pathfinder_deinit(&pf);
  return ret;
}

std::vector<double> GenerateFeatureProbs(
    const std::vector<uint32_t>&    features,
    std::tuple<uint32_t, uint32_t>& block_num,
    double                          false_positive,
    double                          false_negative,
    uint64_t                        seed,
    bool                            normalize) {
  const uint32_t num = std::get<0>(block_num) * std::get<1>(block_num);
  if (num == 0) throw std::runtime_error {"number of blocks is zero"};

  std::vector<double> ret;
  for (auto c : features) {
    const auto begin = ret.size();

    double sum = 0;
    for (uint32_t bi = 0; bi < num; ++bi) {
      seed = blky_numeric_xorshift64(seed);
      const double prob1 = static_cast<double>(seed%1000)/1000.;

      seed = blky_numeric_xorshift64(seed);
      const double prob2 = static_cast<double>(seed%1000)/1000.;

      seed  = blky_numeric_xorshift64(seed);
      double fprob;
      if ((c == bi && prob1 >= false_negative) || prob2 < false_positive) {
        fprob = static_cast<double>(seed%200)/1000. + .8;
      } else {
        fprob = static_cast<double>(seed%800)/1000.;
      }
      ret.push_back(fprob);
      sum += fprob;
    }
    if (normalize) {
      for (size_t i = begin; i < ret.size(); ++i) {
        ret[i] /= sum;
      }
    }
  }
  return ret;
}

}  // namespace blky
