#pragma once

#include <cassert>
#include <cstdint>
#include <cstring>
#include <exception>
#include <fstream>
#include <string_view>
#include <vector>

#include <minimp4.h>

#include <wels/codec_api.h>


class VideoDecoder final {
 public:
  VideoDecoder() = delete;
  VideoDecoder(const std::string& path) :
      file_(path, std::ifstream::binary | std::ifstream::ate),
      size_(static_cast<size_t>(file_.tellg())) {
    // init objects
    SDecodingParam dparam = {};
    dparam.sVideoProperty.eVideoBsType = VIDEO_BITSTREAM_DEFAULT;
    dparam.eEcActiveIdc = ERROR_CON_SLICE_COPY;

    WelsCreateDecoder(&decoder_);
    decoder_->Initialize(&dparam);

    int lv = WELS_LOG_DEBUG;
    decoder_->SetOption(DECODER_OPTION_TRACE_LEVEL, &lv);

    demuxer_ = {};
    MP4D_open(&demuxer_, ReadCallback, this, static_cast<int64_t>(size_));

    // find video track
    track_ = SIZE_MAX;
    for (size_t i = 0; i < demuxer_.track_count; ++i) {
      if (demuxer_.track[i].handler_type == MP4D_HANDLER_TYPE_VIDE) {
        if (track_ != SIZE_MAX) {
          throw std::runtime_error {"there are many video tracks"};
        }
        track_ = i;
      }
    }
    if (track_ == SIZE_MAX) {
      throw std::runtime_error {"there is no video track"};
    }

    // setup decoder
    std::vector<uint8_t> temp_;

    for (size_t i = 0;; ++i) {
      int size;
      auto sps = static_cast<const uint8_t*>(MP4D_read_sps(
              &demuxer_,
              static_cast<unsigned int>(track_),
              static_cast<int>(i),
              &size));
      if (!sps) break;

      temp_.resize(static_cast<size_t>(4+size));
      temp_[0] = 0, temp_[1] = 0, temp_[2] = 0, temp_[3] = 1;
      std::memcpy(&temp_[4], sps, static_cast<size_t>(size));

      if (decoder_->DecodeFrameNoDelay(temp_.data(), static_cast<int>(temp_.size()), yuv_, &frame_)) {
        throw std::runtime_error {"failed to decode SPS"};
      }
    }
    for (size_t i = 0;; ++i) {
      int size;
      auto pps = static_cast<const uint8_t*>(MP4D_read_pps(
              &demuxer_,
              static_cast<unsigned int>(track_),
              static_cast<int>(i),
              &size));
      if (!pps) break;

      temp_.resize(static_cast<size_t>(4+size));
      temp_[0] = 0, temp_[1] = 0, temp_[2] = 0, temp_[3] = 1;
      std::memcpy(&temp_[4], pps, static_cast<size_t>(size));

      if (decoder_->DecodeFrameNoDelay(temp_.data(), static_cast<int>(temp_.size()), yuv_, &frame_)) {
        throw std::runtime_error {"failed to decode SPS"};
      }
    }
    temp_.clear();
  }
  ~VideoDecoder() noexcept {
    decoder_->Uninitialize();
    WelsDestroyDecoder(decoder_);

    MP4D_close(&demuxer_);
  }
  VideoDecoder(const VideoDecoder&) = delete;
  VideoDecoder(VideoDecoder&&) = delete;
  VideoDecoder& operator=(const VideoDecoder&) = delete;
  VideoDecoder& operator=(VideoDecoder&&) = delete;

  void Decode() {
    if (temp_consumed_ >= temp_.size()) {
      unsigned size, time, dur;
      const auto off = MP4D_frame_offset(
          &demuxer_,
          static_cast<unsigned int>(track_),
          static_cast<unsigned int>(count_),
          &size, &time, &dur);

      assert(size > 0);
      temp_.resize(size);
      temp_consumed_ = 0;

      file_.seekg(static_cast<std::streamoff>(off));
      assert(file_);
      file_.read((char*) temp_.data(), size);
      assert(file_);

      Decode();

    } else {
      auto& i = temp_consumed_;

      const uint32_t nal_size = 4 +
          static_cast<uint32_t>((temp_[i+0] << 24) |
                                (temp_[i+1] << 16) |
                                (temp_[i+2] <<  8) |
                                (temp_[i+3] <<  0));
      temp_[i  ] = 0;
      temp_[i+1] = 0;
      temp_[i+2] = 0;
      temp_[i+3] = 1;
      if (decoder_->DecodeFrameNoDelay(temp_.data(), static_cast<int>(nal_size), yuv_, &frame_)) {
        throw std::runtime_error {"failed to decode a frame"};
      }
      i += nal_size;
    }
  }

 private:
  std::ifstream file_;
  size_t size_;

  ISVCDecoder* decoder_;
  MP4D_demux_t demuxer_;

  size_t track_;

  uint8_t*    yuv_[3] = {0};
  SBufferInfo frame_  = {};
  size_t      count_  = 0;

  size_t temp_consumed_ = 0;
  std::vector<uint8_t> temp_;


  static int ReadCallback(int64_t off, void* buf, size_t size, void* ptr) noexcept {
    auto self = (VideoDecoder*) ptr;

    auto n = self->size_ - static_cast<size_t>(off) - size;
    if (size < n) n = size;

    self->file_.seekg(off);
    assert(self->file_);

    self->file_.read((char*) buf, static_cast<std::streamsize>(n));
    assert(self->file_);
    return 0;
  }
};
